from Models import TrainingModel 
import os
from glob import glob
import torchaudio
from torchvision.io import read_image
import torchaudio.transforms as T
import torch
from torchvision import transforms

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

audio_transform = lambda waveform: waveform[:, 32000:]

mfcc= T.MFCC(
        sample_rate=16_000,
        n_mfcc=13,
        log_mels=True,
        melkwargs={"n_fft": 400, "hop_length": 160, "n_mels": 23, "center": False})
image_transform = transforms.Normalize(mean=[0.5, 0.5, 0.5],
                                                    std=[0.5, 0.5, 0.5])
def load_in_sample(mode,path,key):
    img, audio = None, None
    if mode in {"image", "both"}:
        img = read_image(os.path.join(path,key+".png")).float() / 255.0
        img = image_transform(img).to(device)
        img = img.unsqueeze(0)

    if mode in {"audio", "both"}:
        waveform, freq = torchaudio.load(os.path.join(path,key+".wav"))
        assert freq == 16000
        # throw out first two second of silence
        audio = audio_transform(waveform)
        mfccs = mfcc(audio)  # Shape [B, num_mfcc, T]
        mean = mfccs.mean(dim=2, keepdim=True)
        std = mfccs.std(dim=2, keepdim=True)

        audio = (mfccs - mean) / (std + 1e-8)
        audio=audio.to(device)
        audio_length = torch.tensor([[audio.size(2)]]).to(device)

    zeroes_emb = torch.zeros(1,31).to(device)

    if mode == "image":
        return (img, zeroes_emb)
    elif mode == "audio":
        return ((audio, audio_length), zeroes_emb)
    else:
        return ((img, (audio,audio_length)), zeroes_emb)

    

num_classes = 31
fusion = False
mode="image"
path="eval"
output_file="fusion.txt"

# If using only image, only audio, or model that does fusion
# Set path to weights accordingly
if(mode in ["audio","image"] or fusion == True):
    model = TrainingModel(num_classes=31,mode=mode)
    model.load_state_dict(torch.load("resnet-combined-data.pth"))
    model.eval()
    model.to(device)
# If want to use model that uses both image and audio model, and combine probabilities
else:
    model = TrainingModel(num_classes=31,mode="image")
    model.load_state_dict(torch.load("resnet-combined-data.pth"))
    model.eval()
    model.to(device)
    model2 = TrainingModel(num_classes=31,mode="audio")
    model2.load_state_dict(torch.load("x-vector-combined-data.pth"))
    model2.eval()
    model2.to(device)
img_files = sorted(glob(os.path.join(path, "*.png")))
wav_files = sorted(glob(os.path.join(path, "*.wav")))

img_dict = {os.path.splitext(os.path.basename(f))[0]: f for f in img_files}
wav_dict = {os.path.splitext(os.path.basename(f))[0]: f for f in wav_files}

shared_keys = sorted(img_dict.keys())
if img_dict.keys() != wav_dict.keys():
    print(f"Warning: Image samples don't match audio samples")


with open(output_file,"w+") as f:
    with torch.no_grad():
        for key in shared_keys:
            # if(i < start or i >= end):
            #     continue
            data,labels = load_in_sample(mode=mode,path=path,key=key)
            if(mode in ["audio","image"] or fusion==True):
                _,predictions = model(data,labels)
                log_probs = torch.nn.functional.log_softmax(predictions, dim=1)
                ## add one, because classes start from 1
                pred_person = predictions.argmax(dim=1) + 1
            else:
                images,audios = data
                _,predictions = model(images,labels)
                _,predictions2 = model2(audios,labels)

                probs = torch.softmax(predictions, dim=1)     # shape: [1, 31]
                probs2 = torch.softmax(predictions2, dim=1)    # shape: [1, 31]

                combined_probs = probs + probs2
                normalized_probs = torch.softmax(combined_probs, dim=1)

                pred_person = combined_probs.argmax(dim=1) + 1

                # Logarithm of the combined probabilities
                log_probs = torch.log(normalized_probs)
            f.write(key)
            f.write(" ")
            f.write(str(pred_person[0].item()))
            f.write(" ")
            # write log_probs[0] values space-separated
            for val in log_probs[0]:
                f.write(f"{val.item()} ")
            f.write("\n")    

