import torch
from torch.utils.data import DataLoader
from DataLoader import CustomDataset
import torch.nn as nn
import torch.optim as optim
from Models import TrainingModel
from torch.nn.utils.rnn import pad_sequence
import random
# import numpy as np
import torch.nn.functional as F
# import matplotlib.pyplot as plt

def get_device():
    return torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

# def plot_it(train_losses,val_losses,val_accuracies):
#     # Plotting
#     plt.figure(figsize=(12, 5))

#     plt.subplot(1, 2, 1)
#     plt.plot(train_losses, label='Train Loss')
#     plt.plot(val_losses, label='Validation Loss')
#     plt.xlabel('Epoch')
#     plt.ylabel('Loss')
#     plt.title('Loss Curve')
#     plt.legend()

#     plt.subplot(1, 2, 2)
#     plt.plot(val_accuracies, label='Validation Accuracy')
#     plt.xlabel('Epoch')
#     plt.ylabel('Accuracy')
#     plt.title('Validation Accuracy')
#     plt.legend()

#     plt.tight_layout()
#     plt.show()

# nebude fungovat pro both mode
@torch.no_grad()
def validate(model, dataloader, device):
    model.eval()
    total_loss = 0.0
    correct = 0
    total = 0
    model.eval()
    for batch in dataloader:
        inputs, labels = batch
        # if isinstance(inputs, tuple):
        #     inputs = tuple(x.to(device) for x in inputs)
        # else:
        #     inputs = inputs.to(device)
        # labels = labels.to(device)


        loss, logits = model(inputs, labels)
        # loss of batch, is everage loss of a sample, so i multipyl by number of samples
        total_loss += loss.item() * labels.size(0)

        if isinstance(logits, tuple):
            probs = torch.softmax(logits[0], dim=1)     # shape: [1, 31]
            probs2 = torch.softmax(logits[1], dim=1)    # shape: [1, 31]
            logits = probs + probs2 # Added softmaxs together, call them logits :=D


        predictions = logits.argmax(dim=1)
        labels = labels.argmax(dim=1)

        correct += (predictions == labels).sum().item()
        total += labels.size(0)

    avg_loss = total_loss / total
    accuracy = correct / total
    print(f"Validation loss: {avg_loss:.4f}, Accuracy: {accuracy:.2%}")
    return avg_loss,accuracy

def train(model, dataloader, device, epochs, lr,validation_dataloader):
    optimizer = optim.Adam(model.parameters(), lr=lr)
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.8)
    train_losses = []
    val_losses = []
    val_accuracies = []
    for epoch in range(epochs):
        epoch_loss = 0
        model.train() # rather set it manually here, since in validation i turn it of
        for batch in dataloader:
            inputs, labels = batch
            # if isinstance(inputs, tuple):  # (img, audio)
                # inputs = tuple(x.to(device) for x in inputs)
            # else:
            # inputs = inputs.to(device)
            # labels = labels.to(device)

            optimizer.zero_grad()
            loss,_ = model(inputs,labels)
            loss.backward()
            epoch_loss += loss.item()
            optimizer.step()
        avg_train_loss = epoch_loss / len(dataloader)
        print(f"Epoch {epoch + 1},loss {avg_train_loss:.3}.")
        val_loss,val_acc=validate(model, validation_dataloader, device)
        train_losses.append(avg_train_loss)
        val_losses.append(val_loss)
        val_accuracies.append(val_acc)
        scheduler.step()
    # plot_it(train_losses,val_losses,val_accuracies)

def main():
    mode = "both"  # options: "image", "audio", "both"
    training_dataset_path = "train"
    validation_dataset_path = "dev"
    num_classes=31
    batch_size = 16
    epochs = 200
    lr = 0.001
    drop_out = 0.2
    batch_norm = True
    data_augment = True

    # validations_samples = num_classes*2

    # So the results are replicable
    set_seed(42)

    device = get_device()
    print(f"Using device: {device}")

    dataset = CustomDataset(dataset_path=training_dataset_path, mode=mode,num_classes=num_classes,device=device,augment=data_augment)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True,collate_fn=custom_collate_fn)
    validation_dataset = CustomDataset(dataset_path=validation_dataset_path, mode=mode,num_classes=num_classes,device=device)
    validation_dataloader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=False,collate_fn=custom_collate_fn)

    model = TrainingModel(num_classes=num_classes,mode=mode,dropout_p=drop_out,batch_norm=batch_norm)
 

    model.to(device)
    train(model, dataloader, device, epochs, lr, validation_dataloader)
    torch.save(model.state_dict(),"fusion100.pth")

# This collate function paddes the audio, to the length of the longest audio in batch with zeros
def custom_collate_fn(batch):
    # if there is tuple of length two, there is images and audio
    if(len(batch[0][0])) == 2:
        # Unpack batch: each item is ((image, audio), label)
        images, audios, labels = zip(*[(x[0], x[1], y) for (x, y) in batch])

        # Convert all to tensors
        images = torch.stack(images)  # Assuming all images are same shape

        # I have tuple of tensor
        # Each tensor is [1,N]
        # So squeeze the inner tensors, and store in list
        audios = [a.squeeze(0) for a in audios]
        audios = [a.transpose(0, 1) for a in audios]  # now [SEQ_LEN, 13]
        original_lengths = torch.tensor([a.size(0) for a in audios], device=audios[0].device)
        padded = pad_sequence(audios, batch_first=True)
        # Transpose back to [BATCH, 13, MAX_SEQ_LEN]
        padded = padded.transpose(1, 2)

        labels = torch.stack(labels)
        # images is tensort [BATCH,3,80,80]
        # labels is tensort [BATCH,Dimension]
        return (images, (padded,original_lengths)), labels
    # if there are three dimension it is images
    elif(len(batch[0][0]) == 3):
        # Unpack image and label
        images, labels = zip(*[(x, y) for (x, y) in batch])
        # Convert all images to tensors and stack
        images = torch.stack(images)  # Assuming all images are the same shape

        labels = torch.stack(labels)
        # image is tensor [BATCH,3,80,80]
        return images, labels
    else:
        audio, labels = zip(*[(x, y) for (x, y) in batch])
        audios = [a.squeeze(0) for a in audio]
        audios = [a.transpose(0, 1) for a in audios]  # now [SEQ_LEN, 13]
        original_lengths = torch.tensor([a.size(0) for a in audios], device=audios[0].device)
        padded = pad_sequence(audios, batch_first=True)
        # Transpose back to [BATCH, 13, MAX_SEQ_LEN]
        padded = padded.transpose(1, 2)

        labels = torch.stack(labels)
        return (padded, original_lengths), labels


def set_seed(seed):
    random.seed(seed)
    # np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)  # for single-GPU
    torch.cuda.manual_seed_all(seed)  # for multi-GPU

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

if __name__ == "__main__":
    main()
